;    awchunkyfile.pbi
;    Version 0.2 [2013/03/31]
;    Copyright (C) 2011 Ronny Krger
;
;    This file is part of AWPB-Tools.
;
;    AWPB-Tools is free software: you can redistribute it and/or modify
;    it under the terms of the GNU Lesser General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    AWPB-Tools is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU Lesser General Public License for more details.
;
;    You should have received a copy of the GNU Lesser General Public License
;    along with AWPB-Tools.  If not, see <http://www.gnu.org/licenses/>.

XIncludeFile "awzip.pbi"

; Define #AWPB_ChunkyFile_EnableAES if you need encryption support
; #AWPB_ChunkyFile_EnableAES = 1

;{ Constants
#AWPB_FILE_MODE_NONE    = 0
#AWPB_FILE_MODE_READ    = 1
#AWPB_FILE_MODE_WRITE   = 2
;}
;{ support procedures
Procedure.l ____AWPB_MakeHdr(hdr.s)
  Protected h.s, res.l
  
  h   = LSet(hdr, 4, " ")
  res = (Asc(Left(h, 1))) + (Asc(Mid(h, 2, 1)) << 8) + (Asc(Mid(h, 3, 1)) << 16) + (Asc(Mid(h, 4, 1)) << 24)
  
  ProcedureReturn res
EndProcedure
Procedure.s ____AWPB_DecodeHdr(hdr.l)
  Protected res.s
  
  res = Chr((hdr & $000000FF)) + Chr((hdr & $0000FF00) >> 8) + Chr((hdr & $00FF0000) >> 16) + Chr((hdr & $FF000000) >> 24)
  
  ProcedureReturn res
EndProcedure
;}
;{ Structures
Structure ____AWCF_Header
  Header.l
  UserHeader.l
  Flags.l
EndStructure
Structure ____AWCF_ChunkHeader
  Header.l
  Offset.l
  NextChunkOffset.l
  CRC32.l
  DataSize.l  
EndStructure
Structure ____AWCF_CompressedBlockHdr
  OriginalLength.l
  OriginalCRC32.l
  PackedLength.l
  PackedCRC32.l
  Packed.b
  Cipher.b
EndStructure
;}
;{ ChunkyFile-Object
Interface AWPB_ChunkyFile
  ; public
  Destroy               ()
  CloseChunk.b          ()
  Close                 ()
  Open.b                (filename.s, hdr.s, mode.b)
  OpenChunk.b           (hdr.s)
  WriteQuad.b           (value.q)
  WriteLong.b           (value.l)
  WriteWord.b           (value.w)
  WriteByte.b           (value.b)
  WriteDouble.b         (value.d)
  WriteFloat.b          (value.f)
  WriteBlock.b          (*src, slen.i, complevel.i = #AWPB_ZIP_COMP_STORE)
  WriteString.b         (str.s, format.b, complevel.i = #AWPB_ZIP_COMP_STORE)
  NextChunk.b           ()
  GetChunkHeader.s      ()
  GetChunkSize.i        ()
  ReadBlock.i           ()
  GetBlockSize.i        ()
  ReadString.s          ()
  ReadQuad.q            ()
  ReadLong.l            ()
  ReadWord.w            ()
  ReadByte.b            ()
  ReadDouble.d          ()
  ReadFloat.f           ()
  CheckChunkIntegrity.b ()
  FreeBlock             (*block)
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	InitAES.b							(*key, *initializationvector, bits.i = 128)
  	FreeAES								()
  	EnableAES.b						()
  	DisableAES						()
  CompilerEndIf
  
  ; private
  _WriteData.b          (*Binary, BinLen.i)
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	_WriteCipherData.b		(*Binary, BinLen.i)
  	_ReadCipherData.i			(*Buffer, BinLen.i)
  CompilerEndIf

EndInterface
Structure ____AWPB_ChunkyFile
  ; private
  *vt.AWPB_ChunkyFile
  ChunkOffset.i
  Handle.i
  Mode.b
  FileHdr.____AWCF_Header
  ChunkHdr.____AWCF_ChunkHeader
  ChunkOpen.b
  LastBlockLen.i
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	UseAES.b
   	*AESKey
  	*AESInitializationVector
  	AESBits.i
  CompilerEndIf
  
  ; public  
EndStructure
;{ Member Procedures
Procedure.i AWPB_CreateChunkyFileObj()
  Protected *this.____AWPB_ChunkyFile
  
  *this = AllocateMemory(SizeOf(____AWPB_ChunkyFile))
  If *this 
    *this\vt                       = ?____AWPB_ChunkyFile_VT
    *this\Mode                     = #AWPB_FILE_MODE_NONE
    *this\Handle                   = #Null
    *this\ChunkOffset              = -1
    *this\FileHdr\Flags            = 0
    *this\FileHdr\Header           = ____AWPB_MakeHdr("####")
    *this\FileHdr\UserHeader       = ____AWPB_MakeHdr("####")
    *this\ChunkHdr\CRC32           = 0
    *this\ChunkHdr\DataSize        = 0
    *this\ChunkHdr\Header          = ____AWPB_MakeHdr("####")
    *this\ChunkHdr\NextChunkOffset = -1
    *this\ChunkOpen                = #False
    CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
    	*this\UseAES									 = #False
  		*this\AESKey									 = #Null
  		*this\AESBits									 = 0
  		*this\AESInitializationVector	 = #Null
  	CompilerEndIf
  	
    ProcedureReturn *this
  EndIf
  
  ProcedureReturn #Null
EndProcedure
CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
	Procedure.b ____AWPB_ChunkyFile_EnableAES(*this.____AWPB_ChunkyFile)
		If (*this\AESInitializationVector <> #Null) And (*this\AESKey <> #Null)
			*this\UseAES = #True
		Else
			*this\UseAES = #False
		EndIf

		ProcedureReturn *this\UseAES
	EndProcedure
	Procedure   ____AWPB_ChunkyFile_DisableAES(*this.____AWPB_ChunkyFile)
		*this\UseAES = #False
	EndProcedure
	Procedure   ____AWPB_ChunkyFile_FreeAES(*this.____AWPB_ChunkyFile)
		____AWPB_ChunkyFile_DisableAES(*this)
  	If *this\AESInitializationVector : FreeMemory(*this\AESInitializationVector) : EndIf
  	If *this\AESKey									 : FreeMemory(*this\AESKey)									 : EndIf
  	*this\UseAES									= #False
  	*this\AESKey									= #Null
  	*this\AESBits									= 0
  	*this\AESInitializationVector	= #Null  	
	EndProcedure
	Procedure.b ____AWPB_ChunkyFile_InitAES(*this.____AWPB_ChunkyFile, *key, *initializationvector, bits.i = 128)
		____AWPB_ChunkyFile_FreeAES(*this)
		If (*key <> #Null) And (*initializationvector <> #Null) And ((bits = 128) Or (bits = 192) Or (bits = 256))
			*this\AESInitializationVector = AllocateMemory(16)
			If *this\AESInitializationVector <> #Null
				CopyMemory(*initializationvector, *this\AESInitializationVector, 16)
				*this\AESKey = AllocateMemory(bits >> 3)
				If *this\AESKey <> #Null
					CopyMemory(*key, *this\AESKey, bits >> 3)
					*this\AESBits = bits
					*this\UseAES = #False
					
					ProcedureReturn #True
				EndIf
				FreeMemory(*this\AESInitializationVector)
				*this\AESInitializationVector = #Null
			EndIf
		EndIf
		ProcedureReturn #False		
	EndProcedure
CompilerEndIf
Procedure.b ____AWPB_ChunkyFile_CloseChunk(*this.____AWPB_ChunkyFile)
  If *this\ChunkOpen
    If IsFile(*this\Handle)
      *this\ChunkHdr\NextChunkOffset = Loc(*this\Handle)
      FileSeek(*this\Handle, *this\ChunkOffset)
      WriteData(*this\Handle, @*this\ChunkHdr, SizeOf(____AWCF_ChunkHeader))
      FileSeek(*this\Handle, *this\ChunkHdr\NextChunkOffset)
      *this\ChunkOpen = #False
      ProcedureReturn #True
    EndIf
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure   ____AWPB_ChunkyFile_Close(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle)
    If *this\Mode = #AWPB_FILE_MODE_READ
      CloseFile(*this\Handle)
    ElseIf *this\Mode = #AWPB_FILE_MODE_WRITE
      If *this\ChunkOpen : ____AWPB_ChunkyFile_CloseChunk(*this) : EndIf
      CloseFile(*this\Handle)
    EndIf
  EndIf
  *this\Mode                     = #AWPB_FILE_MODE_NONE
  *this\Handle                   = #Null
  *this\ChunkOffset              = -1
  *this\FileHdr\Flags            = 0
  *this\FileHdr\Header           = ____AWPB_MakeHdr("####")
  *this\FileHdr\UserHeader       = ____AWPB_MakeHdr("####")
  *this\ChunkHdr\CRC32           = 0
  *this\ChunkHdr\DataSize        = 0
  *this\ChunkHdr\Header          = ____AWPB_MakeHdr("####")
  *this\ChunkHdr\NextChunkOffset = -1
  *this\ChunkOpen                = #False
EndProcedure
Procedure.b ____AWPB_ChunkyFile_Open(*this.____AWPB_ChunkyFile, filename.s, hdr.s, mode.b) 
  If IsFile(*this\Handle) : ____AWPB_ChunkyFile_Close(*this) : EndIf
  If filename
    *this\Mode = mode
    If mode = #AWPB_FILE_MODE_WRITE
      *this\Handle = CreateFile(#PB_Any, filename)
      If IsFile(*this\Handle)
        *this\FileHdr\Flags            = 0
        *this\FileHdr\Header           = ____AWPB_MakeHdr("AWCF")
        *this\FileHdr\UserHeader       = ____AWPB_MakeHdr(hdr)
        *this\ChunkHdr\CRC32           = 0
        *this\ChunkHdr\DataSize        = 0
        *this\ChunkHdr\Header          = ____AWPB_MakeHdr("####")
        *this\ChunkHdr\NextChunkOffset = -1
        WriteData(*this\Handle, @*this\FileHdr, SizeOf(____AWCF_Header))
        *this\ChunkOffset              = Loc(*this\Handle)
        *this\ChunkOpen                = #False
        ProcedureReturn #True
      EndIf
      *this\Handle = #Null
    ElseIf mode = #AWPB_FILE_MODE_READ
      *this\Handle = ReadFile(#PB_Any, filename)
      If IsFile(*this\Handle)
        If ReadData(*this\Handle, @*this\FileHdr, SizeOf(____AWCF_Header)) = SizeOf(____AWCF_Header)
          If *this\FileHdr\UserHeader = ____AWPB_MakeHdr(hdr)
            If Not Eof(*this\Handle)
              *this\ChunkOffset = Loc(*this\Handle)
              *this\ChunkOpen   = #False
              ProcedureReturn #True
            EndIf
          EndIf
        EndIf
        CloseFile(*this\Handle)
        *this\Handle = #Null
      EndIf
    EndIf
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_ChunkyFile_OpenChunk(*this.____AWPB_ChunkyFile, hdr.s)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_WRITE)
    If *this\ChunkOpen : ____AWPB_ChunkyFile_CloseChunk(*this) : EndIf
    *this\ChunkHdr\Offset          = Loc(*this\Handle)
    *this\ChunkHdr\CRC32           = 0
    *this\ChunkHdr\DataSize        = 0
    *this\ChunkHdr\Header          = ____AWPB_MakeHdr(hdr)
    *this\ChunkHdr\NextChunkOffset = -1
    *this\ChunkOffset              = *this\ChunkHdr\Offset
    *this\ChunkOpen                = #True
    WriteData(*this\Handle, @*this\ChunkHdr, SizeOf(____AWCF_ChunkHeader))
    ProcedureReturn #True
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteData(*this.____AWPB_ChunkyFile, *Binary, BinLen.i)
  If *Binary And IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_WRITE)
    If *this\ChunkOpen
      If *this\ChunkHdr\DataSize
        *this\ChunkHdr\CRC32  = CRC32Fingerprint(*Binary, BinLen, *this\ChunkHdr\CRC32)
      Else
        *this\ChunkHdr\CRC32  = CRC32Fingerprint(*Binary, BinLen)
      EndIf
      *this\ChunkHdr\DataSize + BinLen
      If WriteData(*this\Handle, *Binary, BinLen) = BinLen
      	ProcedureReturn #True
     	EndIf
    EndIf
  EndIf
  
  ProcedureReturn #False
EndProcedure
CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
	Procedure.b ____AWPB_ChunkyFile_WriteCipherData(*this.____AWPB_ChunkyFile, *Binary, BinLen.i)
		Protected *dst, *src, srclen.i, res.b
		
		res = #False
		If *Binary And IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_WRITE)
			If BinLen >= 16
				*src   = *Binary
				srclen = BinLen
			Else
				*src   = AllocateMemory(16)
				srclen = 16
				If *src <> #Null
					FillMemory(*src, 16)
					CopyMemory(*Binary, *src, BinLen)	
				EndIf
			EndIf
			If *src
				*dst = AllocateMemory(srclen)
				If *dst
					If AESEncoder(*src, *dst, srclen, *this\AESKey, *this\AESBits, *this\AESInitializationVector) <> 0
						res = ____AWPB_ChunkyFile_WriteData(*this, *dst, srclen)
					EndIf
					FreeMemory(*dst)
				EndIf
				If *src <> *Binary
					FreeMemory(*src)
				EndIf
			EndIf
		EndIf
		
	  ProcedureReturn res
	EndProcedure
CompilerEndIf
Procedure.b ____AWPB_ChunkyFile_WriteQuad(*this.____AWPB_ChunkyFile, value.q)
	ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteLong(*this.____AWPB_ChunkyFile, value.l)
  ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteWord(*this.____AWPB_ChunkyFile, value.w)
  ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteByte(*this.____AWPB_ChunkyFile, value.b)
  ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteDouble(*this.____AWPB_ChunkyFile, value.d)
  ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteFloat(*this.____AWPB_ChunkyFile, value.f)
  ProcedureReturn ____AWPB_ChunkyFile_WriteData(*this, @value, SizeOf(value))
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteBlock(*this.____AWPB_ChunkyFile, *src, slen.i, complevel.i)
  Protected hdr.____AWCF_CompressedBlockHdr, *dst, dlen.i, scrc.l, dcrc.l
  
  If *src And slen
    If *this\ChunkOpen And IsFile(*this\Handle)
      dlen = 0
      *dst = #Null
      scrc = CRC32Fingerprint(*src, slen)   ; Prfsumme der Originaldaten berechnen
      If complevel < #AWPB_ZIP_COMP_NONE : complevel = #AWPB_ZIP_COMP_NONE : EndIf
      If complevel > #AWPB_ZIP_COMP_MAX  : complevel = #AWPB_ZIP_COMP_MAX  : EndIf
      If complevel
        *dst = ____ZIP_PackMemory(*src, slen, @dlen, complevel)
      EndIf
      If (dlen = 0) Or (dlen > slen)         ; wenn nicht gepackt oder nach dem packen grer
        hdr\OriginalCRC32  = scrc
        hdr\OriginalLength = slen
        hdr\PackedCRC32    = scrc
        hdr\PackedLength   = slen
        hdr\Packed         = 0
				CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
					If *this\UseAES : hdr\Cipher = 1 : Else : hdr\Cipher = 0 : EndIf
				CompilerElse
					hdr\Cipher = 0
				CompilerEndIf
        ____AWPB_ChunkyFile_WriteData(*this, @hdr, SizeOf(____AWCF_CompressedBlockHdr))
				CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
					If *this\UseAES
						____AWPB_ChunkyFile_WriteCipherData(*this, *src, slen)  ; data
					Else
						____AWPB_ChunkyFile_WriteData(*this, *src, slen)  ; data
					EndIf
				CompilerElse
					____AWPB_ChunkyFile_WriteData(*this, *src, slen)  ; data
				CompilerEndIf
      Else                                                ; erfolgreich komprimiert
        dcrc = CRC32Fingerprint(*dst, dlen)               ; Prfsumme der gepackten Daten berechnen
        hdr\OriginalCRC32  = scrc
        hdr\OriginalLength = slen
        hdr\PackedCRC32    = dcrc
        hdr\PackedLength   = dlen
        hdr\Packed         = 1
				CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
					If *this\UseAES : hdr\Cipher = 1 : Else : hdr\Cipher = 0 : EndIf
				CompilerElse
					hdr\Cipher = 0
				CompilerEndIf
        ____AWPB_ChunkyFile_WriteData(*this, @hdr, SizeOf(____AWCF_CompressedBlockHdr))
				CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
					If *this\UseAES
		        ____AWPB_ChunkyFile_WriteCipherData(*this, *dst, dlen)  ; data
					Else
		        ____AWPB_ChunkyFile_WriteData(*this, *dst, dlen)  ; data
					EndIf
				CompilerElse
	        ____AWPB_ChunkyFile_WriteData(*this, *dst, dlen)  ; data
				CompilerEndIf
      EndIf
    
      If *dst : FreeMemory(*dst) : EndIf
      
      ProcedureReturn #True 
    EndIf
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_ChunkyFile_WriteString(*this.____AWPB_ChunkyFile, str.s, format.b, complevel.i)
  Protected *buf, len.i, res.b, blen.i
  
  res = #False
  If str
    ; Speicher anfordern fr den String im gewnschten Format und anschlieend inkl. Chr(0)-Abschluss in die Datei schreiben, falls gewnscht gepackt
    len  = Len(str)
    blen = StringByteLength(str, format) + 2
    *buf = AllocateMemory(blen)
    If *buf
      PokeS(*buf, str, len, format)
      PokeW(*buf + blen - 2, 0)
      res = ____AWPB_ChunkyFile_WriteLong(*this, blen)
      res & ____AWPB_ChunkyFile_WriteByte(*this, format)
      res & ____AWPB_ChunkyFile_WriteBlock(*this, *buf, blen, complevel)
      FreeMemory(*buf)
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.b ____AWPB_ChunkyFile_NextChunk(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And (*this\ChunkOffset > 0)
    FileSeek(*this\Handle, *this\ChunkOffset)
    If Not Eof(*this\Handle)
      ReadData(*this\Handle, @*this\ChunkHdr, SizeOf(____AWCF_ChunkHeader))
      *this\ChunkOffset = *this\ChunkHdr\NextChunkOffset
      *this\ChunkOpen   = #True   
      ProcedureReturn #True
    EndIf
  EndIf
  *this\ChunkOffset = -1
  *this\ChunkOpen   = #False
  ProcedureReturn #False
EndProcedure
Procedure.s ____AWPB_ChunkyFile_GetChunkHeader(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ____AWPB_DecodeHdr(*this\ChunkHdr\Header)
  EndIf
  
  ProcedureReturn ""
EndProcedure
Procedure.i ____AWPB_ChunkyFile_GetChunkSize(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn *this\ChunkHdr\DataSize
  EndIf
  
  ProcedureReturn -1
EndProcedure
CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
	Procedure.i ____AWPB_ChunkyFile_ReadCipherData(*this.____AWPB_ChunkyFile, *Buffer, BinLen.i)
		Protected *dst, *src, srclen.i, res.b
		
		res = #False
		If IsFile(*this\Handle) And (*Buffer <> #Null) And (BinLen > 0)
			If BinLen >= 16
				*src   = AllocateMemory(BinLen)
				*dst   = *Buffer
				srclen = BinLen
			Else
				*src   = AllocateMemory(16)
				*dst   = AllocateMemory(16)
				srclen = 16
			EndIf
			If (*src <> #Null) And (*dst <> #Null)
				If ReadData(*this\Handle, *src, srclen) = srclen
					If AESDecoder(*src, *dst, srclen, *this\AESKey, *this\AESBits, *this\AESInitializationVector) <> 0
						If *dst <> *Buffer : CopyMemory(*dst, *Buffer, BinLen) : EndIf
						res = #True
					EndIf
				EndIf
			EndIf
			If *dst <> *Buffer
				If *dst <> #Null : FreeMemory(*dst) : EndIf
			EndIf
			If *src <> #Null : FreeMemory(*src) : EndIf
		EndIf
		
	  ProcedureReturn res
	EndProcedure
CompilerEndIf
Procedure.i ____AWPB_ChunkyFile_ReadBlock(*this.____AWPB_ChunkyFile)
  Protected hdr.____AWCF_CompressedBlockHdr, *src, *dst, slen.i, dlen.i, packed.b, pos.i, len.i, scrc.l, dcrc.l, crc.l, rlen.i
  
  *this\lastblocklen = -1
  *dst               = #Null
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ReadData(*this\Handle, @hdr, SizeOf(____AWCF_CompressedBlockHdr))  
    If hdr\Packed
      dlen = hdr\OriginalLength
      slen = hdr\PackedLength
      dcrc = hdr\OriginalCRC32
      scrc = hdr\PackedCRC32
      *src = AllocateMemory(slen)
      If *src
      	If hdr\Cipher = 1
      		CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)      	
      			If *this\UseAES
      				If ____AWPB_ChunkyFile_ReadCipherData(*this, *src, slen)
      					rlen = slen
      				Else 
      					rlen = 0
      				EndIf
      			Else
      				rlen = 0
      			EndIf
      		CompilerElse
      			rlen = 0
      		CompilerEndIf
      	Else
      		rlen = ReadData(*this\handle, *src, slen)
      	EndIf
      	
      	If rlen = slen
          crc = CRC32Fingerprint(*src, slen)
          If crc = scrc
            *dst = ____ZIP_UnpackMemory(*src, slen, @len)
            FreeMemory(*src)
            *src = #Null
          EndIf
        EndIf
      EndIf
    Else
      dlen = hdr\OriginalLength
      dcrc = hdr\OriginalCRC32
      *dst = AllocateMemory(dlen)
      *src = #Null
      If *dst
      	If hdr\Cipher = 1
      		CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)      	
      			If *this\UseAES
      				If ____AWPB_ChunkyFile_ReadCipherData(*this, *dst, dlen)
      					rlen = dlen
      				Else 
      					rlen = 0
      				EndIf
      			Else
      				rlen = 0
      			EndIf
      		CompilerElse
      			rlen = 0
      		CompilerEndIf
      	Else
      		rlen = ReadData(*this\handle, *dst, dlen)
      	EndIf
      	
      	If rlen = dlen
          len = dlen
        EndIf
      EndIf
    EndIf
    
    If len = dlen
      crc = CRC32Fingerprint(*dst, dlen)
      If crc = dcrc
        *this\lastblocklen = len
      Else
        FreeMemory(*dst)
        *dst = #Null
      EndIf
    Else
      FreeMemory(*dst)
      *dst = #Null
    EndIf
  EndIf
  
  ProcedureReturn *dst
EndProcedure
Procedure.i ____AWPB_ChunkyFile_GetBlockSize(*this.____AWPB_ChunkyFile)
  ProcedureReturn *this\LastBlockLen
EndProcedure
Procedure.s ____AWPB_ChunkyFile_ReadString(*this.____AWPB_ChunkyFile)
  Protected res.s, format.b, slen.i, *src
  
  res = ""
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    slen   = ReadLong(*this\Handle)
    format = ReadByte(*this\Handle)
    *src   = ____AWPB_ChunkyFile_ReadBlock(*this)
    If *src 
      If slen
        res = PeekS(*src, slen, format)
      EndIf
      FreeMemory(*src)
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.q ____AWPB_ChunkyFile_ReadQuad(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadQuad(*this\handle)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.l ____AWPB_ChunkyFile_ReadLong(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadLong(*this\handle)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.w ____AWPB_ChunkyFile_ReadWord(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadWord(*this\handle)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.b ____AWPB_ChunkyFile_ReadByte(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadByte(*this\handle)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.d ____AWPB_ChunkyFile_ReadDouble(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadDouble(*this\handle)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.f ____AWPB_ChunkyFile_ReadFloat(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    ProcedureReturn ReadFloat(*this\handle)
  EndIf

  
  ProcedureReturn 0
EndProcedure
Procedure.b ____AWPB_ChunkyFile_CheckChunkIntegrity(*this.____AWPB_ChunkyFile)
  Protected p.i, *src, scrc.l, res.b
  
  res = #False
  If IsFile(*this\Handle) And (*this\mode = #AWPB_FILE_MODE_READ) And *this\ChunkOpen
    p = Loc(*this\Handle)
    FileSeek(*this\Handle, *this\ChunkHdr\Offset + SizeOf(____AWCF_ChunkHeader))
    *src = AllocateMemory(*this\ChunkHdr\DataSize)
    If *src
      ReadData(*this\Handle, *src, *this\ChunkHdr\DataSize)
      scrc = CRC32Fingerprint(*src, *this\ChunkHdr\DataSize)  
      FreeMemory(*src)
      If scrc = *this\ChunkHdr\CRC32
        res = #True
      EndIf
    EndIf
    FileSeek(*this\Handle, p)
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure   ____AWPB_ChunkyFile_FreeBlock(*this.____AWPB_ChunkyFile, *block)
  If *block
    FreeMemory(*block)
  EndIf
EndProcedure
Procedure.i ____AWPB_ChunkyFile_Destroy(*this.____AWPB_ChunkyFile)
  If IsFile(*this\Handle) : ____AWPB_ChunkyFile_Close(*this) : EndIf
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	____AWPB_ChunkyFile_FreeAES(*this)
  CompilerEndIf
  FreeMemory(*this)  
  ProcedureReturn #Null
EndProcedure
;}
DataSection ;{ Virtual Function Table
____AWPB_ChunkyFile_VT:
  ; public
  Data.i @____AWPB_ChunkyFile_Destroy()
  Data.i @____AWPB_ChunkyFile_CloseChunk()
  Data.i @____AWPB_ChunkyFile_Close()
  Data.i @____AWPB_ChunkyFile_Open()
  Data.i @____AWPB_ChunkyFile_OpenChunk()
  Data.i @____AWPB_ChunkyFile_WriteQuad()
  Data.i @____AWPB_ChunkyFile_WriteLong()
  Data.i @____AWPB_ChunkyFile_WriteWord()
  Data.i @____AWPB_ChunkyFile_WriteByte()
  Data.i @____AWPB_ChunkyFile_WriteDouble()
  Data.i @____AWPB_ChunkyFile_WriteFloat()
  Data.i @____AWPB_ChunkyFile_WriteBlock()
  Data.i @____AWPB_ChunkyFile_WriteString()
  Data.i @____AWPB_ChunkyFile_NextChunk()
  Data.i @____AWPB_ChunkyFile_GetChunkHeader()
  Data.i @____AWPB_ChunkyFile_GetChunkSize()
  Data.i @____AWPB_ChunkyFile_ReadBlock()
  Data.i @____AWPB_ChunkyFile_GetBlockSize()
  Data.i @____AWPB_ChunkyFile_ReadString()
  Data.i @____AWPB_ChunkyFile_ReadQuad()
  Data.i @____AWPB_ChunkyFile_ReadLong()
  Data.i @____AWPB_ChunkyFile_ReadWord()
  Data.i @____AWPB_ChunkyFile_ReadByte()
  Data.i @____AWPB_ChunkyFile_ReadDouble()
  Data.i @____AWPB_ChunkyFile_ReadFloat()
  Data.i @____AWPB_ChunkyFile_CheckChunkIntegrity()
  Data.i @____AWPB_ChunkyFile_FreeBlock()
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	Data.i @____AWPB_ChunkyFile_InitAES()
  	Data.i @____AWPB_ChunkyFile_FreeAES()
  	Data.i @____AWPB_ChunkyFile_EnableAES()
  	Data.i @____AWPB_ChunkyFile_DisableAES()
  CompilerEndIf
  
  ; private
  Data.i @____AWPB_ChunkyFile_WriteData()
  CompilerIf Defined(AWPB_ChunkyFile_EnableAES, #PB_Constant)
  	Data.i @____AWPB_ChunkyFile_WriteCipherData()
  	Data.i @____AWPB_ChunkyFile_ReadCipherData()
  CompilerEndIf
  
EndDataSection ;}
;}

 
; IDE Options = PureBasic 5.11 (Linux - x64)
; CursorPosition = 745
; Folding = AgAAAwAA-
; EnableXP